@@ -1,7 +1,6 @@
 import ast
 import itertools
 import random
-
 from ast import literal_eval as leval
 from copy import deepcopy
 from typing import List, Union
@@ -11,15 +10,16 @@ import pandas as pd
 from casadi.casadi import ceil, floor
 from rdkit import Chem, RDLogger, rdBase
 from rdkit.Chem import AllChem, Descriptors
-from rdkit.Chem.rdchem import ChiralType
-
 from tqdm import tqdm
 
+from .stereo_reactions import polyester_stereo, polyvinyl_stereo
+
 tqdm.pandas()
 lg = RDLogger.logger()
 # lg.setLevel(RDLogger.ERROR)
 lg.setLevel(4)
 
+
 class PolyMaker:
     def __init__(self):
         self.smiles_req = {
@@ -31,7 +31,6 @@ class PolyMaker:
             "cyclic_carbonates": "[O]=[C]1[O][C][C][O]1",
             "acidanhydrides": "[#8]([#6](=[#8]))([#6](=[#8]))",
             "prime_thiols": "[#6;!$(C=O)][SH]",
-            "isocyanates":"[#6][#7;!$([#7+])]=[#6]=[#8]"
         }
         self.reactions = {
             "ester": {
@@ -85,17 +84,10 @@ class PolyMaker:
                 "[C:7][NH:6][C:1](=[O:0])[O:2][C:3][C:4][O:5]",
                 "diamine_cycliccarbonate": "[NH2;!$([NH2+]):6][#6;!$(C=O):7].[O:0]=[C:1]1[O:2][C:3][C:4][O:5]1>>"
                 "[C:7][NH:6][C:1](=[O:0])[O:2][C:3][C:4][O:5]",
-                "infinite_chain": "to complete",
+                "infinite_chaine": "to complete",
             },
-            "Urethane":{
-                "diisocyanates_ol":"[#6:0][#7;!$([#7+]):1]=[#6:2]=[#8:3].[C,c;!$(C=O):4][OH:5]>>"
-                "[#6:0][#7;!$([#7+]):1]-[#6:2](=[#8:3])[O:5][C;!$(C=O):4]",
-                "diols_isocyanate":"[C,c;!$(C=O):4][OH:5].[#6:0][#7;!$([#7+]):1]=[#6:2]=[#8:3]>>"
-                "[#6:0][#7;!$([#7+]):1]-[#6:2](=[#8:3])[O:5][C;!$(C=O):4]",
-                "infinite_chain":"to complete"
-            }
         }
-        self.__verison__ = "0.1.4.4"
+        self.__verison__ = "0.1.5.0"
 
     @staticmethod
     def checksmile(smi: str) -> Union[str, None]:
@@ -121,7 +113,7 @@ class PolyMaker:
         return smi
 
     @staticmethod
-    def get_monomers(smi: str, stereochemistry: bool = False) -> List[str]:
+    def get_monomers(smi: str, stereochemistry: bool = True) -> List[str]:
         """Convert a string of monomers into a list of monomers.
 
         The input string must contain monomer SMILES seperated by ".". Each monomer is checked for validity.
@@ -143,19 +135,29 @@ class PolyMaker:
         except:
             pass
 
-        if type(smi) == str:
-            smi = smi.split(".")
-            if not stereochemistry:
-                smi = [smi_i.replace("/", "").replace("@", "") for smi_i in smi]
-            monomers = tuple([PolyMaker.checksmile(smi_i) for smi_i in smi])
-            if np.any(
-                np.array(monomers) == "",
-            ):
-                monomers == None
-        elif type(smi) == tuple:
-            monomers = smi
+        # Determine whether or no enantiomers are present
+        if ".." not in smi:
+            if type(smi) == str:
+                smi = smi.split(".")
+                if not stereochemistry:
+                    smi = [smi_i.replace("/", "").replace("@", "") for smi_i in smi]
+                monomers = tuple([PolyMaker.checksmile(smi_i) for smi_i in smi])
+            elif type(smi) == tuple:
+                monomers = smi
+            else:
+                monomers = None
         else:
-            monomers = None
+            # assume for now there are monomers with R and S and specified like
+            # M1R.M1S..M2R.M2S
+            monomers = [
+                [PolyMaker.checksmile(enant) for enant in enants.split(".")]
+                for enants in smi.split("..")
+                if enants
+            ]
+
+        # Check to see if there are any bad smiles in flattened list
+        if any([item == "" for sublist in monomers for item in sublist]):
+            monomers == None
 
         return monomers
 
@@ -292,7 +294,6 @@ class PolyMaker:
         self,
         reactants: pd.DataFrame,
         DP: int = 2,
-        mechanism: str = "",
         replicate_structures: int = 1,
         distribution: Union[List[float], List] = [],
         pm: float = 1,
@@ -303,15 +304,10 @@ class PolyMaker:
         Parameters
         ----------
         reactants : pd.DataFrame
-            A pandas dataframe containing a smiles column, a monomers column (generated by get_monomers), and an option pm column
+            A pandas dataframe containing a smiles column, a monomers column (generated by get_monomers),
+            a mechanism column, an optional pm column, and an optinional distribution column
         DP : int, optional
             Degree of polymerization, by default 2
-        mechanism : str, optional
-            Polymerization mechanism
-                vinyl: performs polymerization along vinyl groups
-                ester: performs condensation reaction on dicarboxylic acid + diol
-                amide: performs condensation reaction on dicarboxylic acid + diamine
-                carbonate: performs condensation reaction on carbonate + diol
             by default ''
         replicate_structures : int, optional
             Number of replicate structures per monomer entry, by default 1
@@ -327,217 +323,89 @@ class PolyMaker:
         pd.DataFrame
             A dataframe containing the polymerized inputs
         """
+        # Get rid of these two different ones. Using now just for getting it to work.
 
-        def get_RS_assignments(n_structures, pm):
-            """Generate a a list containing the R/S assignemnts for compounds. First calculates diads then converts to R/S."""
-            # Initiate info to generate a diad list
-            n_diads = DP - 1
-            lower, upper = [int(floor(n_diads * pm)), int(ceil(n_diads * pm))]
-            diad_lists = [
-                np.array([1] * lower + [0] * (n_diads - lower)),
-                np.array([1] * upper + [0] * (n_diads - upper)),
-            ]
-
-            # Get weights of sampling so average pm (with infinite replicates) is equal to specified pm
-            if lower != upper:
-                r_weight = (n_diads * pm - lower / n_diads) / (upper - lower)
-                l_weight = 1 - r_weight
-                weights = [l_weight, r_weight]
-            else:
-                weights = [1, 1]
-
-            # Generate n_structures amount of diad lists
-            # Ideally unique, but only try a max iteration of 10^3
-            i = 0
-            replicates = []
-            max_iter = 10 ** 3
-            while len(replicates) < n_structures:
-                # Generate a random permutation and convert it into a string
-                diad_list = random.choices(diad_lists, weights)[0]
-                perm = tuple(np.random.permutation(diad_list))
-                perm_string = "".join([f"{num}" for num in perm])
-
-                # Add permutation to replicate list and reset iteration count
-                if (perm_string not in replicates) or (i >= max_iter):
-                    replicates.append(perm_string)
-                    i = 0
-
-                i += 1
-
-            # Convert to R/S with arbitrary starting stereo for first atom
-            RS = ["R", "S"]
-            RS_replicates = []
-            for diads in replicates:
-                current_RS = np.random.randint(0, 2)
-                RS_list = [RS[current_RS % 2]]
-
-                for diad in diads:
-                    if diad == "0":
-                        current_RS += 1
-                    RS_list.append(RS[current_RS % 2])
-                RS_replicates.append(RS_list)
-
-            return RS_replicates
-
-        def get_stereo_smiles_from_CIP(RS_list, smi, chiral_centers):
-            """Assign stereochemistry from a list of CIP values."""
-
-            def get_incorrect_chirality(known_chirality, mol_chirality):
-                """Get the ids of incorrect atom chirality"""
-                incorrect_set = set(known_chirality.items()).symmetric_difference(
-                    mol_chirality
-                )
-                incorrect_atom_set = set()
-                for mol_chirality in incorrect_set:
-                    incorrect_atom_set.add(mol_chirality[0])
-                return list(incorrect_atom_set)
-
-            chiral_dict = {
-                "R": ChiralType.CHI_TETRAHEDRAL_CCW,
-                "S": ChiralType.CHI_TETRAHEDRAL_CW,
-            }
-            known_chirality = {key: val for key, val in zip(*(chiral_centers, RS_list))}
-
-            # Set values with R as CCW, S as CW. This isn't always true, but we fix later
-            # This assumes there is one pendant group, can fail
-            smi = Chem.CanonSmiles(smi)
-            mol = Chem.MolFromSmiles(smi)
-            atoms = mol.GetAtoms()
-            for atom_i, CIP_id in known_chirality.items():
-                atoms[atom_i].SetChiralTag(chiral_dict[CIP_id])
-            AllChem.AssignCIPLabels(mol)
-
-            # Identify any incorrect centers and flip them
-            current_chirality = AllChem.FindMolChiralCenters(mol, includeCIP=True)
-            incorrect_chirality = get_incorrect_chirality(
-                known_chirality, current_chirality
-            )
-            atoms = mol.GetAtoms()
-
-            for atom_i in incorrect_chirality:
-                current_chirality = atoms[atom_i].GetChiralTag()
-                if current_chirality == ChiralType.CHI_TETRAHEDRAL_CCW:
-                    atoms[atom_i].SetChiralTag(ChiralType.CHI_TETRAHEDRAL_CW)
-                else:
-                    atoms[atom_i].SetChiralTag(ChiralType.CHI_TETRAHEDRAL_CCW)
-
-            AllChem.AssignCIPLabels(mol)
-
-            return Chem.MolToSmiles(mol)
-
-        def generate_stereo_ester(smi, n_structures, pm):
-            """
-            Generate stereo structures for duplicate smiles of a polyester.
-            This can fail when there are multiple stereo sites in a monomer backbone.
-            """
-            # Get stereo centers first
-            smi = Chem.CanonSmiles(smi)
-            mol = AllChem.MolFromSmiles(smi)
-            acid = AllChem.MolFromSmarts("[C;$(C[OH]);$(C=O)]")
-            ol = AllChem.MolFromSmarts("[C;$(C[OH]);!$(C=O)]")
-
-            acid_atoms = mol.GetSubstructMatches(acid)
-            ol_atoms = mol.GetSubstructMatches(ol)
-
-            chiral_centers = AllChem.FindMolChiralCenters(mol, includeUnassigned=True)
-            chiral_centers = [center[0] for center in chiral_centers]
-
-            # Get backbone and backbone chiral sites
-            shortest_path = set(
-                AllChem.rdmolops.GetShortestPath(
-                    mol,
-                    ol_atoms[0][0],
-                    acid_atoms[0][0],
-                )
-            )
-            chiral_centers = [
-                center for center in chiral_centers if center in shortest_path
-            ]
-
-            # Get diads for given pm
-            replicate_RS = get_RS_assignments(n_structures, pm)
-            smiles_list = []
-            for RS_list in replicate_RS:
-                smiles_list.append(
-                    get_stereo_smiles_from_CIP(RS_list, smi, chiral_centers)
-                )
-
-            return smiles_list
+        df_return_mapped = pd.DataFrame()
 
         # Ensure there is a Pm value, default is 1
         if "pm" not in reactants:
             reactants["pm"] = pm
         reactants[["pm"]] = reactants[["pm"]].fillna(value=pm)
 
-        # if method == "classic":
-        #     # Find duplicates based on polymer, pm, and mechanism and generate an R/S list for these
-        #     # Assumes no duplicates in the input
-        #     staging_df = pd.DataFrame()
-        #     # duplicates_df = staging_df.groupby(by=["pm", "smiles"]).size().to_frame('n_rows').reset_index()
-        #     for _, row in reactants.iterrows():
-        #         RS_assignments = ["".join(RS_str) for RS_str in get_RS_assignments(replicate_structures, pm)]
-        #         smiles_df = pd.DataFrame({"RS_assignment": RS_assignments})
-        #         smiles_df[row.index] = row.values
-        #         smiles_df = pd.DataFrame(data={"smiles": smiles})
-        #         poly_df.loc[(poly_df.smiles_polymer == row.smiles_polymer) & (poly_df.pm == row.pm), "RS_assignment"] = get_RS_assignments(replicate_structures, pm)
-        # else:
-        # Don't want stereochemistry on the first pass for synthesis
-        nostereo_df = reactants.copy()
-        nostereo_df.loc[:, "smiles_monomer"] = nostereo_df.smiles_monomer.map(
-            lambda smiles: ".".join(
-                [
-                    smi.replace("@", "")
-                    for i, smi in enumerate(smiles.split("."))
-                    if i % 2 == 0
-                ]
+        if "distribution" in reactants:
+            reactants["distribution"] = reactants["distribution"].fillna(
+                str(distribution)
             )
-        )
-        nostereo_df.loc[:, "monomers"] = nostereo_df.monomers.map(
-            lambda monomers: [
-                monomer.replace("@", "")
-                for i, monomer in enumerate(monomers)
-                if i % 2 == 0
+            reactants["distribution"] = reactants["distribution"].astype(str)
+        else:
+            reactants.loc[:, "distribution"] = str(distribution)
+
+        if "replicate_structures" in reactants:
+            reactants["replicate_structures"] = reactants[
+                "replicate_structures"
+            ].fillna(int(replicate_structures))
+        else:
+            reactants["replicate_structures"] = int(replicate_structures)
+
+        reactants["DP"] = DP
+
+        stereo_df = reactants.copy()
+        if not stereo_df.empty:
+            stereo_df.loc[:, "monomers"] = stereo_df.monomers.astype(str)
+            df_return_mapped = pd.concat([df_return_mapped, stereo_df])
+
+        if verbose:
+            df_return_mapped = df_return_mapped.progress_apply(
+                self._polymerizemechanism_thermoplastic_stereo, axis=1,
+            )
+        else:
+            df_return_mapped = df_return_mapped.apply(
+                self._polymerizemechanism_thermoplastic_stereo, axis=1,
+            )
+
+        if not df_return_mapped.empty:
+            df_return_mapped = pd.concat([res for res in df_return_mapped]).reset_index(
+                drop=True
+            )
+
+            # if not df_return_mapped.empty:
+            #     df_return_mapped.sort_index().sort_values("replicate_structure")
+
+        if "index" in df_return_mapped.columns:
+            df_return_mapped = df_return_mapped.drop(columns=["index"])
+        df_return_mapped = df_return_mapped.drop(columns=["replicate_structures"])
+        df_return_mapped = df_return_mapped.reset_index(drop=True)
+
+        return df_return_mapped
+
+    def _polymerizemechanism_thermoplastic_stereo(self, row):
+        df_poly = pd.DataFrame()
+        if (row.mechanism == "vinyl_stereo") | (row.mechanism == "acrylate_stereo"):
+            df_poly = polyvinyl_stereo(
+                row.monomers,
+                row.DP,
+                row.replicate_structures,
+                row.distribution,
+                row.pm,
+            )
+        elif row.mechanism == "ester_stereo":
+            df_poly = polyester_stereo(
+                row.monomers,
+                row.DP,
+                row.replicate_structures,
+                row.distribution,
+                row.pm,
+            )
+
+        if isinstance(df_poly, str) or df_poly.empty:
+            # TODO: how to integrate
+            return pd.DataFrame()
+        else:
+            df_poly[row.keys()] = row
+            df_poly["replicate_structure"] = [
+                i for i in range(row.replicate_structures)
             ]
-        )
-        poly_df = self.thermoplastic(
-            nostereo_df,
-            DP,
-            mechanism,
-            replicate_structures,
-            distribution,
-            False,
-            verbose,
-        )
-
-        # Find duplicates based on polymer, pm, and mechanism and generate the stereo for these
-        duplicates_df = (
-            poly_df.groupby(by=["pm", "smiles_polymer"])
-            .size()
-            .to_frame("n_rows")
-            .reset_index()
-        )
-        for _, row in duplicates_df.iterrows():
-            smiles = generate_stereo_ester(row.smiles_polymer, row.n_rows, row.pm)
-            smiles_df = pd.DataFrame(data={"smiles_monomer": smiles})
-            poly_df.loc[
-                (poly_df.smiles_polymer == row.smiles_polymer) & (poly_df.pm == row.pm),
-                "smiles_polymer",
-            ] = smiles_df.smiles_monomer.to_numpy()
-
-        poly_df["index"] = poly_df.index
-        reactants["index"] = reactants.index
-
-        for _, row in reactants.iterrows():
-            poly_df.loc[
-                poly_df["index"] == row["index"], "smiles_monomer"
-            ] = row.smiles_monomer
-            poly_df.loc[poly_df["index"] == row["index"], "monomers"] = str(
-                row.monomers
-            )
-
-        poly_df = poly_df.drop(columns=["index"])
-
-        return poly_df
+            return df_poly
 
     def thermoplastic(
         self,
@@ -677,11 +545,6 @@ class PolyMaker:
                     Chem.MolFromSmarts(self.smiles_req["cyclic_carbonates"])
                 )
             )
-            r.isocyanates = len(
-                mol.GetSubstructMatches(
-                    Chem.MolFromSmarts(self.smiles_req["isocyanates"])
-                )
-            )
             return r
 
         df_func = pd.DataFrame(
@@ -695,7 +558,6 @@ class PolyMaker:
                 "aliphatic_ols",
                 "acidanhydrides",
                 "cyclic_carbonates",
-                "isocyanates"
             ],
         )
         df_func = df_func.apply(lambda r: id_functionality(r), axis=1)
@@ -791,11 +653,6 @@ class PolyMaker:
             polydata = self.__poly_NIPU(reactants, DP, distribution, infinite_chain)
             returnpoly = polydata[0]
             mechanism = polydata[1]
-        
-        elif mechanism == "urethane":
-            polydata = self.__poly_urethane(reactants, DP, distribution, infinite_chain)
-            returnpoly = polydata[0]
-            mechanism = polydata[1]
 
         elif mechanism == "all":
             polylist = [
@@ -805,7 +662,6 @@ class PolyMaker:
                 self.__poly_carbonate(reactants, DP, distribution, infinite_chain),
                 self.__poly_imide(reactants, DP, distribution, infinite_chain),
                 self.__poly_NIPU(reactants, DP, distribution, infinite_chain),
-                self.__poly_urethane(reactants, DP, distribution, infinite_chain),
             ]
 
             polylist = [
@@ -819,7 +675,6 @@ class PolyMaker:
                     "ERROR:Carbonate_ReactionFailed",
                     "ERROR:Imide_ReactionFailed",
                     "ERROR:NIPU_ReactionFailed",
-                    "ERROR:Urethane_ReactionFailed",
                     "",
                 ]
             ]
@@ -1011,260 +866,6 @@ class PolyMaker:
             polymer = "ERROR:Vinyl_ReactionFailed"
         return polymer, "vinyl"
 
-    def __poly_ester_stereo(
-        self, reactants, DP=2, pm=0.5, distribution=[1], infinite_chain=False
-    ):
-        """performs condenstation reaction on dicarboxyl and  diols
-        A poly ester generator that incorporates stereochemistry. There are three polymer properties that are controlled:
-                (1) DP, degree of polymerization. Counts each monomer provided as a repeat unit (each adds 1 to DP)
-                (2) pm, which is the probability for meso addition. 0 = syndiotactic, 0.5 = atactic, 1 = isotactic
-                (3) the distribution of monomers in the polymer (fraction composition)
-        In order to properly construct the polyer, the inputs must follow the following conventions:
-                (1) reactants must be smiles in a tuple. Unique monomer order is determined by stripping stereochemistry
-                        from monomers. For each monomer it is assumed that two provided monomers are the R and S enantiomers in that
-                        order.
-                (2) pm must be a single value, which is applied to all monomers equally, or a list that corresponds to the
-                        order of monomers in the reactants tuple.
-                (3) distribution must be a single value, which is applied to all monomers equally, or a list that
-                        corresponds to the order of the monomers in the reactants tuple.
-        Example inputs, where monomer letters represent their SMILES representation.
-                (1) reacting one monomer, A, with two enantiomers, A(R) and A(S) in an isotactic fashion (pm=1).
-                                reactants = (A(R), A(S))
-                                pm = 1
-                (2) reacting two monomers: A, with two enantiomers, A(R) and A(S), and B, with two enatniomers,
-                        B(R) and B(S), with A adding isotactically and B adding atactically, and a ratio of A:B of 80:20
-                                reactants = (A(R), A(S), B(R), B(S))
-                                pm = [1, 0.5]
-                                distribution = [80, 20]
-                (3) reacting two monomers: A, with two enantiomers, A(R) and A(S) and B, with no enantiomers.
-                        A is syndiotactic. A pm value must be supplied to every monomer, even if it can't display tacticity.
-                        This can be done with either a list, or just using one pm value.
-                                reactants = (A(R), A(S), B)
-                                pm = 0
-        reactants: Tuple[str]
-                A tuple containing the reactant smiles
-        DP: int
-                Degree of polymerization. Each monomer added contributes to this value, by default 2
-        pm: Union[float, List[float]]
-                The probability for meso addition for the monomer sepcies. 0 = syndiotactic, 0.5 = atactic, 1 = isotactic.
-                Specifying one value sets the value for all monomers, by default 0.5
-        distribution: List[float]
-                The distribution for the monomer species in the resultant polymer. Specifying one value sets the value for
-                all monomers, by default [1]
-        infinite_chain: bool
-                Whether or not to use an infinite chain, by default False
-        """
-        # Define helper functions
-        def replace_acidanhydrides(reactant):
-            mol = Chem.MolFromSmiles(reactant)
-            if (
-                len(
-                    mol.GetSubstructMatches(
-                        Chem.MolFromSmarts(self.smiles_req["acidanhydrides"])
-                    )
-                )
-                > 0
-            ):
-                reactant = self.__openacidanyhydride(reactant)
-            else:
-                pass
-            return reactant
-
-        def sample_by_pm(monomers, last_stereo):
-            """Sample monomer to react based on pm values
-            monomers: df
-                    Dataframe containing monomers to sample from
-            last_stereo: int
-                    The stereochemistry of the last monomer
-            """
-            # Select id of monomer to add
-            monomer_id = sample_monomer_id(
-                distribution, ids=monomers.monomer_id.unique()
-            )
-
-            # Get any monomers that match
-            monomers = monomers[
-                (monomers.index != "smiles_polymer")
-                & (monomers["monomer_id"] == monomer_id)
-            ].reset_index(drop=False)
-
-            if last_stereo == 2:  # No stereochemistry, sample randomly
-                new_monomer = monomers.sample(1)
-            elif len(monomers) == 1:  # Only one monomer to select
-                new_monomer = monomers
-            else:
-                # Generate weights based on pm and get sample monomer
-                monomer_pm = pm[monomer_id]
-                if last_stereo == 0:
-                    weights = [monomer_pm, 1 - monomer_pm]
-                else:
-                    weights = [1 - monomer_pm, monomer_pm]
-
-                new_monomer = monomers.sample(1, weights=weights)
-
-            # Return new monomer SMILES and new stereo_id
-            return new_monomer.iloc[0]["index"], new_monomer.iloc[0]["stereo_id"]
-
-        def sample_monomer_id(distribution, ids=[]):
-            # Get a monomer ID from the distribution
-            # ids dictates which polymers to draw from, empty list indicates all
-            p_distribution = np.array(distribution)
-            ids = np.array(ids)
-
-            if len(ids) > 0:
-                p_distribution = p_distribution[ids]
-            else:
-                ids = range(len(distribution))
-
-            # normalize for the choice function
-            p_distribution = np.array(p_distribution, dtype=float)
-            p_distribution /= sum(p_distribution)
-
-            monomer_id = np.random.choice(ids, p=p_distribution)
-
-            return monomer_id
-
-        try:
-            # open acid anhydrides
-            reactants = pd.Series(reactants).apply(replace_acidanhydrides).tolist()
-
-            # Load reaction info and get functionalities
-            rxn_dict = self.reactions["ester"]
-            func_df = self.get_functionality(reactants)
-
-            # Assign each monomer a unique ID independent of stereochemistry to aid in sampling
-            func_df["nonstereo_smiles"] = func_df.index.map(
-                lambda s: s.replace("@", "").replace("\\", "")
-            )
-            func_df["monomer_id"] = func_df.groupby(
-                ["nonstereo_smiles"], sort=False
-            ).ngroup()
-            n_unique_monomers = max(func_df["monomer_id"]) + 1
-
-            # if distribution is wrong shape, make uniform
-            if len(distribution) != n_unique_monomers:
-                distribution = [1] * n_unique_monomers
-
-            # Verify pm is correct shape
-            if type(pm) != list:
-                pm = [pm] * n_unique_monomers
-            else:
-                assert len(pm) == len(reactants)
-
-            # Give each monomer a stereo ID, assumes R is first and S is second
-            # R = 0, S = 1, one monomer = 2
-            for monomer_id in set(func_df["monomer_id"]):
-                n_monomers = len(func_df[func_df["monomer_id"] == monomer_id])
-                assert n_monomers < 3  # Only allowing for 2 monomers currently
-                if n_monomers == 2:
-                    func_df.loc[func_df["monomer_id"] == monomer_id, "stereo_id"] = [
-                        0,
-                        1,
-                    ]
-                elif n_monomers == 1:
-                    func_df.loc[func_df["monomer_id"] == monomer_id, "stereo_id"] = 2
-
-            func_df = func_df.drop(columns=["nonstereo_smiles"])
-
-            # Grab a random polymer from func_df based on the distribution and initialize poly entry in df
-            poly_df = func_df[
-                func_df["monomer_id"] == sample_monomer_id(distribution)
-            ].sample(1)
-            molpoly = Chem.MolFromSmiles(poly_df.index[0])  # index is SMILES
-            poly_df.index.values[0] = "smiles_polymer"
-            func_df = pd.concat([func_df, poly_df])
-            last_stereo = poly_df["stereo_id"][0]
-
-            DP_count = 1
-            DP_actual = 1
-
-            while DP_count < DP:
-                # select rxn rule and reactant
-                if (func_df.loc["smiles_polymer", "aliphatic_ols"] >= 1) & (
-                    func_df.loc["smiles_polymer", "acids"] >= 1
-                ):
-                    mask = ((func_df.acids >= 1) | (func_df.aliphatic_ols >= 1)) & (
-                        func_df.index != "smiles_polymer"
-                    )
-                    func_df_select = func_df.loc[mask]
-                    monomer, last_stereo = sample_by_pm(func_df_select, last_stereo)
-                    if func_df.loc[monomer].aliphatic_ols >= 1:
-                        rxn_selector = "diacids_ols"
-                    if func_df.loc[monomer].acids >= 1:
-                        rxn_selector = "diols_acids"
-
-                elif func_df.loc["smiles_polymer", "aliphatic_ols"] >= 2:
-                    msk = (func_df.acids >= 1) & (func_df.index != "smiles_polymer")
-                    func_df_select = func_df.loc[msk]
-                    monomer, last_stereo = sample_by_pm(func_df_select, last_stereo)
-                    rxn_selector = "diols_acids"
-
-                elif func_df.loc["smiles_polymer", "acids"] >= 2:
-                    msk = (func_df.aliphatic_ols >= 1) & (
-                        func_df.index != "smiles_polymer"
-                    )
-                    func_df_select = func_df.loc[msk]
-                    monomer, last_stereo = sample_by_pm(func_df_select, last_stereo)
-                    rxn_selector = "diacids_ols"
-
-                else:
-                    assert False
-
-                rxn = Chem.AllChem.ReactionFromSmarts(rxn_dict[rxn_selector])
-                # update func_df table
-                func_df.loc["smiles_polymer"] = (
-                    func_df.loc["smiles_polymer"] + func_df.loc[monomer]
-                )  # adding polymer and a
-                for column_name in ["aliphatic_ols", "ols", "acids"]:
-                    func_df.loc[
-                        "smiles_polymer", column_name
-                    ] += -1  # substracting off functionality
-                assert (
-                    func_df.loc["smiles_polymer"][
-                        func_df.loc["smiles_polymer"] > -1
-                    ].shape
-                    == func_df.loc["smiles_polymer"].shape
-                )
-
-                # React and select product
-                mola = Chem.MolFromSmiles(monomer)
-                prod = rxn.RunReactants((molpoly, mola))
-                prodlist = [Chem.MolToSmiles(x[0]) for x in prod]
-                prodlist = self._returnvalid(prodlist)
-                poly = random.choice(prodlist)
-                molpoly = Chem.MolFromSmiles(poly)
-
-                # manage loop and ring close
-                if (infinite_chain) & (DP_count == DP - 1):
-                    # logic for closing ring
-                    if (func_df.loc["smiles_polymer", "aliphatic_ols"] > 0) & (
-                        func_df.loc["smiles_polymer", "acids"]
-                    ) > 0:
-                        # case for when has can ring close
-                        DP_count += 1
-                        DP_actual += 1
-                    else:
-                        # case for when has same terminal ends so can't ring close
-                        DP_count = DP_count
-                        DP_actual += 1
-                else:
-                    DP_count += 1
-                    DP_actual += 1
-
-            if infinite_chain:  # closes ring
-                rxn = Chem.AllChem.ReactionFromSmarts(rxn_dict["infinite_chain"])
-                prod = rxn.RunReactants((molpoly,))
-                prodlist = [Chem.MolToSmiles(x[0]) for x in prod]
-                prodlist = self._returnvalid(prodlist)
-                poly = random.choice(prodlist)
-                molpoly = Chem.MolFromSmiles(poly)
-
-        except BaseException as e:
-            # print(e) # Can give reason why it fails?
-            poly = "ERROR:Ester_Stereo_ReactionFailed"
-
-        return poly, "ester_stereo"
-
     def __protect_substructure(self, mol, substructure, n_unprotected=0):
         """protects atoms in the group identified
 
@@ -1918,116 +1519,6 @@ class PolyMaker:
 
         return poly, "NIPU"
 
-    def __poly_urethane(self, reactants, DP=2, distribution=[], infinite_chain=False):
-        """performs condenstation reaction on diisocyanates and  diols"""
-        # function
-
-        try:
-            # initial
-            rxn_dic = self.reactions["Urethane"]
-            df_func = self.get_functionality(reactants, distribution=distribution)
-
-            # select initial monomer as polymer chain
-            df_poly = df_func.sample(1)
-            df_func.loc["smiles_polymer"] = df_poly.sample(1).values[0]
-            poly = df_poly.index[0]
-            molpoly = Chem.MolFromSmiles(poly)
-
-            DP_count = 1
-            DP_actual = 1
-
-            while DP_count < DP:
-
-                # select rxn rule and reactant
-                if (df_func.loc["smiles_polymer", "isocyanates"] >= 1) & (
-                    df_func.loc["smiles_polymer", "ols"] >= 1
-                ):
-                    msk = (
-                        (df_func.ols >= 1) | (df_func.isocyanates >= 1)
-                    ) & (df_func.index != "smiles_polymer")
-                    df_func_select = df_func.loc[msk]
-                    a = df_func_select.sample(
-                        1, weights=df_func.distribution, replace=True
-                    ).index.values[0]
-                    if df_func.loc[a].isocyanates >= 1:
-                        rxn_selector = "diols_isocyanate"
-                    if df_func.loc[a].ols >= 1:
-                        rxn_selector = "diisocyanates_ol"
-                elif df_func.loc["smiles_polymer", "isocyanates"] >= 2:
-                    msk = (df_func.ols >= 1) & (
-                        df_func.index != "smiles_polymer"
-                    )
-                    df_func_select = df_func.loc[msk]
-                    a = df_func_select.sample(
-                        1, weights=df_func.distribution, replace=True
-                    ).index.values[0]
-                    rxn_selector = "diisocyanates_ol"
-                elif df_func.loc["smiles_polymer", "ols"] >= 2:
-                    msk = (df_func.isocyanates >= 1) & (
-                        df_func.index != "smiles_polymer"
-                    )
-                    df_func_select = df_func.loc[msk]
-                    a = df_func_select.sample(
-                        1, weights=df_func.distribution, replace=True
-                    ).index.values[0]
-                    rxn_selector = "diols_isocyanate"
-                else:
-                    assert False
-
-                rxn = Chem.AllChem.ReactionFromSmarts(rxn_dic[rxn_selector])
-
-                # update df_func table
-                df_func.loc["smiles_polymer"] = (
-                    df_func.loc["smiles_polymer"] + df_func.loc[a]
-                )  # adding polymer and a
-                for column_name in ["isocyanates", "ols"]:
-                    df_func.loc["smiles_polymer", column_name] += -1
-                assert (
-                    df_func.loc["smiles_polymer"][
-                        df_func.loc["smiles_polymer"] > -1
-                    ].shape
-                    == df_func.loc["smiles_polymer"].shape
-                )
-
-                # React and select product
-                mola = Chem.MolFromSmiles(a)
-                prod = rxn.RunReactants((molpoly, mola))
-                prodlist = [Chem.MolToSmiles(x[0]) for x in prod]
-                prodlist = self._returnvalid(prodlist)
-                poly = random.choice(prodlist)
-                molpoly = Chem.MolFromSmiles(poly)
-
-                # manage loop and ring close
-                if (infinite_chain) & (DP_count == DP - 1):
-                    # logic for closing ring
-                    if (df_func.loc["smiles_polymer", "isocyanates"] > 0) & (
-                        df_func.loc["smiles_polymer", "ols"]
-                    ) > 0:
-                        # case for when can ring close
-                        DP_count += 1
-                        DP_actual += 1
-                    else:
-                        # case for when has same terminal ends so can't ring close
-                        DP_count = DP_count
-                        DP_actual += 1
-                else:
-                    DP_count += 1
-                    DP_actual += 1
-
-            if infinite_chain:  # closes ring
-
-                rxn = Chem.AllChem.ReactionFromSmarts(rxn_dic["infinite_chain"])
-                prod = rxn.RunReactants((molpoly,))
-                prodlist = [Chem.MolToSmiles(x[0]) for x in prod]
-                prodlist = self._returnvalid(prodlist)
-                poly = random.choice(prodlist)
-                molpoly = Chem.MolFromSmiles(poly)
-        except:
-            poly = "ERROR:Urethane_ReactionFailed"
-
-        return poly, "Urethane"
-
-
     def __poly_upe(self, reactants, crosslinker, distribution, DP):
         """generates 2 ringed thermoset
         reactants: list of smiles
